#include <stdlib.h>
#include <stdio.h>
#include <chrono>
#include <cuda_runtime.h>
void check(cudaError_t result, char const *const func, const char *const file,
            int const line) {
    if (result) {
        fprintf(stderr, "CUDA error at %s:%d code=%d(%s) \"%s\" \n", file, line,
                 static_cast<unsigned int>(result), cudaGetErrorName(result), func);
        exit(EXIT_FAILURE);
    }
}


#define checkCudaErrors(val) check((val), #val, __FILE__, __LINE__)

__device__ __host__ float add( float a, float b )
{
    return a + b;
}

// Device kernel code
__global__ void VecAdd(float* A, float* B, float* C, int N)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < N)
        C[i] = add(A[i], B[i]);
}



// Host code
int main()
{
    int N = 1024 << 15;
    size_t size = N * sizeof(float);

    // Allocate input vectors h_A and h_B in host memory
    float* h_A = (float*)malloc(size);
    float* h_B = (float*)malloc(size);
    float* h_C = (float*)malloc(size);
    float* h_Cref = (float*)malloc(size);

    // Initialize input vectors
    // fill h_A and h_B with random floats
    for ( int i = 0; i < N; i++ )
    {
        h_A[ i ] = rand() / (float)RAND_MAX;
        h_B[ i ] = rand() / (float)RAND_MAX;
    }

    // Compute reference solution

    std::chrono::time_point<std::chrono::high_resolution_clock> startHost, stopHost;

    startHost = std::chrono::high_resolution_clock::now();
    for ( int i = 0; i < N; i++ )
    {
        h_Cref[i] = add(h_A[i], h_B[i]);
    }
    stopHost = std::chrono::high_resolution_clock::now();

    std::chrono::duration<double> elapsedHost = stopHost - startHost;
    printf( "Time to compute on host:  %3.1f ms\n", elapsedHost.count() * 1000 );

    // Allocate vectors in device memory
    float* d_A;
    checkCudaErrors( cudaMalloc( &d_A, size ) );
    float* d_B;
    checkCudaErrors( cudaMalloc( &d_B, size ) );
    float* d_C;
    checkCudaErrors( cudaMalloc( &d_C, size ) );

    float elapsedTime = 0.0f;

    // create events

    cudaEvent_t startMemcpyUp, stopMemcpyUp, start, stop, startMemcpyDwn, stopMemcpyDwn;
    checkCudaErrors( cudaEventCreate( &startMemcpyUp ) );
    checkCudaErrors( cudaEventCreate( &stopMemcpyUp ) );
    checkCudaErrors( cudaEventCreate( &start ) );
    checkCudaErrors( cudaEventCreate( &stop ) );
    checkCudaErrors( cudaEventCreate( &startMemcpyDwn ) );
    checkCudaErrors( cudaEventCreate( &stopMemcpyDwn ) );

    checkCudaErrors( cudaEventRecord( startMemcpyUp, 0 ) );
    // Copy vectors from host memory to device memory
    checkCudaErrors( cudaMemcpy( d_A, h_A, size, cudaMemcpyHostToDevice ) );
    checkCudaErrors( cudaMemcpy( d_B, h_B, size, cudaMemcpyHostToDevice ) );

    checkCudaErrors( cudaEventRecord( stopMemcpyUp, 0 ) );

    checkCudaErrors( cudaEventRecord( start ) );

    // Invoke kernel
    int threadsPerBlock = 256;
    int blocksPerGrid =
        (N + threadsPerBlock - 1) / threadsPerBlock;

    VecAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, N);
    checkCudaErrors( cudaGetLastError() );
    checkCudaErrors( cudaEventRecord( stop, 0 ) );

    checkCudaErrors( cudaEventRecord( startMemcpyDwn, 0 ) );
    // Copy result from device memory to host memory
    // h_C contains the result in host memory
    checkCudaErrors( cudaMemcpy( h_C, d_C, size, cudaMemcpyDeviceToHost ) );
    checkCudaErrors( cudaEventRecord( stopMemcpyDwn, 0 ) );

    checkCudaErrors( cudaEventSynchronize( stopMemcpyDwn ) );

    checkCudaErrors( cudaEventElapsedTime( &elapsedTime, startMemcpyUp, stopMemcpyUp ) );

    printf( "Time to copy TO device:  %3.1f ms\n", elapsedTime );

    checkCudaErrors( cudaEventElapsedTime( &elapsedTime, start, stop ) );
    printf( "Time to compute:  %3.1f ms\n", elapsedTime );

    checkCudaErrors( cudaEventElapsedTime( &elapsedTime, startMemcpyDwn, stopMemcpyDwn ) );
    printf( "Time to copy FROM device:  %3.1f ms\n", elapsedTime );

    // Free events
    checkCudaErrors( cudaEventDestroy( startMemcpyUp ) );
    checkCudaErrors( cudaEventDestroy( stopMemcpyUp ) );
    checkCudaErrors( cudaEventDestroy( start ) );
    checkCudaErrors( cudaEventDestroy( stop ) );
    checkCudaErrors( cudaEventDestroy( startMemcpyDwn ) );
    checkCudaErrors( cudaEventDestroy( stopMemcpyDwn ) );
    // Free device memory
    checkCudaErrors( cudaFree(d_A) );
    checkCudaErrors( cudaFree(d_B) );
    checkCudaErrors( cudaFree(d_C) );

    // Free host memory
    free( h_A );
    free( h_B );
    free( h_C );
    free( h_Cref );
}
